/**************************************************
opengate Mac addr auth program

 module to control ip addresss cache 

  As ip address check by database is time consuming procedure,
  the recently checked addresses are cached and skiped.
  Implemented with HashTable and Queue.
  HashTable has recentry checked ipAddress and checked time.
  If ip is included in table and time is new, ignore checking.
  Queue has ipAddresses odrered by checked time.
  If an old item is found in table, elder items are removed from table.
  The queue controls the remove sequence.


Copyright (C) 2011 Opengate Project Team
Written by Yoshiaki Watanabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: watanaby@is.saga-u.ac.jp
**************************************************/
#include "opengatemd.h"

#define CACHESIZE 1000000 /* mac&ip cache size */

int InitQueueForCache(void);
int EnQueueForCache(unsigned char* addrRaw, int addrLen);
int DeQueueForCache(unsigned char* addrRaw, int* pAddrLen);
int ListQueueForCache(void);
void FreeQueueForCache(void);

/* Queue to store MacAndIpAddress */
struct queueNode{
  int addrLen;
  unsigned char addrRaw[MACADDRLN+IPV6ADDRLN];
  struct queueNode *next;
};
static struct queueNode* queueTail=NULL;
static struct queueNode* queueHead=NULL;

/* HashTable to store MacAndIpAddress and Time */
static DB* hashDb;

/* Cache Timeout(seconds) = DB checking interval */
static int cacheTimeout;

static int cacheItemCount=0;

/**********************************
This cache is made from HashTable and Queue.
HashTabel for quick access, and Queue for ordering.
Same data are stored in the two data structures.
If you add/delete items in cache, treat both structures.
Don't add/delete items from one structure only. 
***********************************/

/****************************************
Is the IpAddress checked recently or not
 input=ipAddress return TRUE if checked recently
****************************************/
int isRecentlyCheckedAddress(unsigned char* macAndIpAddressRaw, int addrLen){

  int timeNow;
  unsigned char storedAddrRaw[MACADDRLN+IPV6ADDRLN];
  int storedAddrLen;
  int ret;
  int* pTime;
  DBT hashKey;
  DBT hashVal;

  /* get present time */
  timeNow=time(NULL);

  /***** get item matched to the mac&ip from hash table */
  hashKey.data = macAndIpAddressRaw;
  hashKey.size = addrLen;
  memset(&hashVal, 0, sizeof(DBT));
  ret=hashDb->get(hashDb, &hashKey, &hashVal, 0);

  /* get success */
  if(ret==0){

    /* if data is null, return not checked */
    pTime=(int*)(hashVal.data);
    if(pTime==NULL) return FALSE;

    /* if ignoring addresses, return true */
    if(*pTime==0) return TRUE;

    /* if recently checked data, return true */
    if( (timeNow-*pTime) < cacheTimeout ) return TRUE;

    /***** if timeover, elder items are removed from queue and hashTable */
    /* get data from queue head and inspect the address */
    /* until present address is found. */
    while(DeQueueForCache(storedAddrRaw, &storedAddrLen)){
      hashKey.data=storedAddrRaw;
      hashKey.size = storedAddrLen;
      hashDb->del(hashDb, &hashKey, 0);
      if(memcmp(macAndIpAddressRaw,storedAddrRaw,storedAddrLen)==0)break;
    }

    /* insert update item to queue and hashTable */
    EnQueueForCache(storedAddrRaw, storedAddrLen);
    hashVal.data = &timeNow;
    hashVal.size = sizeof(int);    
    if(hashDb->put(hashDb, &hashKey, &hashVal, 0) == -1) {
      err_msg("ERR at %s#%d: fail to put into hash table",__FILE__,__LINE__);
      terminateProg(0);
    }

    /* and return false */
    return FALSE;
  }

  /* if getting from hash fails, insert address and return false */
  if(ret==1){

    /*************** begin adding item to Ip Cache ***/
    /* insert to hash table */
    hashVal.data = &timeNow;
    hashVal.size = sizeof(int);    
    if(hashDb->put(hashDb, &hashKey, &hashVal, 0) == -1) {
      err_msg("ERR at %s#%d: fail to put into hash table",__FILE__,__LINE__);
      terminateProg(0);
    }
    /* insert to queue */
    EnQueueForCache(macAndIpAddressRaw, addrLen);
    /*************** end adding item to Ip Cache ***/

    /* if cache size is over, remove oldest one */
    if(cacheItemCount>CACHESIZE) DelOldestIpCacheItem();
    return FALSE;
  }

  /* else error exit */
  err_msg("ERR at %s#%d: fail to get hash table item",__FILE__,__LINE__);
  return FALSE;
}

/****************************************
initialize IpAddress Cache
 Ip Cache is formed with in HashTable and Queue
 HashTable=to search an item quickly
 Queue    =to list items in FIFO order
****************************************/
void initIpCache(void) {

  /* prepare hash table */
  if((hashDb = dbopen(NULL, O_CREAT | O_RDWR, 0644, DB_HASH, NULL)) == NULL) {
    err_msg("ERR at %s#%d: fail to open hash table",__FILE__,__LINE__);
    terminateProg(0);
  }

  /* prepare queue */
  InitQueueForCache();

  /* set timeout parameter */
  cacheTimeout=atoi(GetConfValue("CacheTimeout"));
}

/****************************************
memory free for IpAddress Cache
****************************************/
void freeIpCache(void) {

  hashDb->close(hashDb);
  FreeQueueForCache();
}

/****************************************
delete item from MacAndIp cache matched to the mac address
****************************************/
int delIpCacheItem(char* macAddress) {
  int found=FALSE;
  DBT hashKey;
  unsigned char addrRaw[MACADDRLN+IPV6ADDRLN];
  struct queueNode *temp;
  struct queueNode *prev;
  
  /* if null or illegal form, return */
  if(isNull(macAddress)) return FALSE;
  if(!ReFormatMacAddr(macAddress)) return FALSE;

  /* make network raw mac address */
  memcpy(addrRaw, ether_aton(macAddress), MACADDRLN);

  /* scan queue to find matched mac address */
  /* point to head of mac-ip cache */
  if(queueHead==NULL) return FALSE;
  prev=queueHead;
  temp=queueHead->next;

  /* scan from queue head to tail */
  while(temp->next!=NULL){

    /* if MAC address part(first 6bytes) is matched to the argument value */
    if(memcmp(addrRaw, temp->addrRaw, MACADDRLN)==0){

      /* set found flag */
      found=TRUE;

      /* delete the item from Hash Table */
      hashKey.data = temp->addrRaw;
      hashKey.size = temp->addrLen;
      hashDb->del(hashDb, &hashKey, 0);

      /* delete the item from Queue */
      prev->next=temp->next;
      free(temp);
      temp=prev;
    }

    /* move to next item */
    prev=temp;
    temp=temp->next;
  }
  return found;
}

/****************************************
delete oldest item from MacAndIp cache 
****************************************/
int delOldestIpCacheItem(void) {
  DBT hashKey;
  unsigned char addrRaw[MACADDRLN+IPV6ADDRLN];
  int addrLen=0;
  
  /* delete oldest item(=head) from queue */
  if(DeQueueForCache(addrRaw, &addrLen)){

    /* delete the item from Hash Table */
    hashKey.data = addrRaw;
    hashKey.size = addrLen;
    hashDb->del(hashDb, &hashKey, 0);
    
    return TRUE;
  }

  return FALSE;
}

/*********************************************
initialize MacAndIpAddress Queue
 Queue
 HeadNode - DataNode - DataNode - TailNode
 (dummy)                           (dummy)
  ^queueHead                        ^queueTail
*********************************************/
int initQueueForCache(void){

  unsigned char addrRaw[MACADDRLN+IPV6ADDRLN];
  int addrLen;

  /* if not exist, prepare head and tail */
  if(queueHead==NULL){
    queueHead=(struct queueNode*)malloc(sizeof(struct queueNode));
    if(queueHead==NULL){
      err_msg("ERR at %s#%d: fail to malloc",__FILE__,__LINE__);
      terminateProg(0);
    }
    queueTail=(struct queueNode*)malloc(sizeof(struct queueNode));
    if(queueTail==NULL){
      err_msg("ERR at %s#%d: fail to malloc",__FILE__,__LINE__);
      terminateProg(0);
    }
    bzero(queueHead->addrRaw, MACADDRLN+IPV6ADDRLN);
    queueHead->addrLen=0;
    bzero(queueTail->addrRaw, MACADDRLN+IPV6ADDRLN);
    queueTail->addrLen=0;
    queueHead->next=queueTail;
    queueTail->next=NULL;
  }
  
  /* if exist, reset all */
  else{
    while(DeQueueForCache(addrRaw,&addrLen))
      ;
  }

  /* reset item count */
  cacheItemCount=0;

  return TRUE;
}

/****************************************
Add data to the tail of MacAndIP Queue
 input=addr
****************************************/
int enQueueForCache(unsigned char* addrRaw, int addrLen){
  struct queueNode *newNode;

  /* if not prepared, error */
  if(queueHead==NULL){
    err_msg("ERR at %s#%d: queue not init",__FILE__,__LINE__);
    return FALSE;
  }

  /* add item after the tail and set it as new tail*/
  newNode=(struct queueNode*)malloc(sizeof(struct queueNode));
  if(newNode==NULL){
    err_msg("ERR at %s#%d: fail to malloc",__FILE__,__LINE__);
    terminateProg(0);
  }
  memcpy(queueTail->addrRaw, addrRaw, addrLen);
  queueTail->addrLen=addrLen;
  queueTail->next=newNode;
  queueTail=newNode;
  bzero(queueTail->addrRaw,MACADDRLN+IPV6ADDRLN);
  queueTail->addrLen=0;
  queueTail->next=NULL;

  /* increment item count */
  cacheItemCount++;

  return TRUE;
}

/****************************************
Get and remove address data from the head of MacAndIP Queue
output
 addrRaw:binary string of Mac&Ip (length=MACADDRLN+IPV6ADDRLN)
 pAddrLen: pointer to the aquired string length 
****************************************/
int deQueueForCache(unsigned char* addrRaw, int* pAddrLen){

  /* set null string as default */
  bzero(addrRaw, MACADDRLN+IPV6ADDRLN);

  /* if not prepared, error */
  if(queueHead==NULL){
    err_msg("ERR at %s#%d: queue not init",__FILE__,__LINE__);
    return FALSE;
  }
  else if(queueHead->next==NULL){
    err_msg("ERR at %s#%d: queue not init",__FILE__,__LINE__);
    return FALSE;  
  }

  /* if no data, return false */
  else if(queueHead->next==queueTail){
    return FALSE;
  }

  /* get item from the head */
  else {
    struct queueNode *temp;
    temp=queueHead->next;
    queueHead->next=temp->next;
    memcpy(addrRaw, temp->addrRaw, temp->addrLen);
    *pAddrLen=temp->addrLen;
    free(temp);

    /* decrement item count */
    cacheItemCount--;
  }

  return TRUE;
}

/****************************************
Listing MacAndIpAddress Queue (for debugging)
****************************************/
int listQueueForCache(void){

  struct queueNode *temp;
  int i;

  if(queueHead==NULL) return FALSE;
  temp=queueHead->next;
  while(temp->next!=NULL){
    for(i=0;i<temp->addrLen;i++){
      printf("%x ", temp->addrRaw[i]);
    }
    printf("\n");
    temp=temp->next;
  }
  return TRUE;
}

/****************************************
memory free for MacAndIpAddress Queue
****************************************/
void freeQueueForCache(void){
  unsigned char addrRaw[MACADDRLN+IPV6ADDRLN];
  int addrLen;
  while(DeQueueForCache(addrRaw,&addrLen));
  free(queueHead);
  free(queueTail);
}

/************************************/
/* ndp form is reformed to arp form */ 
/* format macAddr for ndp or arp    */
/* match the form of two program    */
/* mac addr by arp 00:01:12:0b:..   */
/* mac addr by ndp 0:1:12:b:..      */
/************************************/
int reFormatMacAddr(char* macAddr)
{
  int m1,m2,m3,m4,m5,m6;
  if(sscanf(macAddr, "%x:%x:%x:%x:%x:%x", &m1,&m2,&m3,&m4,&m5,&m6)!=6) return FALSE;
  snprintf(macAddr, ADDRMAXLN,"%02x:%02x:%02x:%02x:%02x:%02x", m1,m2,m3,m4,m5,m6);
  return TRUE;
}


/****************************************************
 routines for debugging putput
 ***************************************************/

void InitIpCache(void) {
  if(debug>1) err_msg("DEBUG:=>initIpCache( )");
  initIpCache();
  if(debug>1) err_msg("DEBUG:<=initIpCache( )");
}

void FreeIpCache(void) {
  if(debug>1) err_msg("DEBUG:=>freeIpCache()");
  freeIpCache();
  if(debug>1) err_msg("DEBUG:<=freeIpCache()");
}

int IsRecentlyCheckedAddress(unsigned char* macAndIpAddressRaw, int addrLen){
  int ret;
  if(debug>1) err_msg("DEBUG:=>isRecentlyCheckedAddress(%x,%d)", macAndIpAddressRaw[0],addrLen);
  ret = isRecentlyCheckedAddress(macAndIpAddressRaw, addrLen);
  if(debug>1) err_msg("DEBUG:(%d)<=isRecentlyCheckedAddress( )",ret);
  return ret;
}

int DelIpCacheItem(char* macAddress) {
  int ret;
  if(debug>1) err_msg("DEBUG:=>delIpCacheItem(%s)", macAddress);
  ret = delIpCacheItem(macAddress);
  if(debug>1) err_msg("DEBUG:(%d)<=delIpCacheItem( )",ret);
  return ret;
}

int DelOldestIpCacheItem(void) {
  int ret;
  if(debug>1) err_msg("DEBUG:=>delOldestIpCacheItem( )");
  ret = delOldestIpCacheItem();
  if(debug>1) err_msg("DEBUG:(%d)<=delOldestIpCacheItem( )",ret);
  return ret;
}

int InitQueueForCache(void){
  int ret;
  if(debug>1) err_msg("DEBUG:=>initQueueForCache( )");
  ret = initQueueForCache();
  if(debug>1) err_msg("DEBUG:(%d)<=initQueueForCache( )",ret);
  return ret;
}

int EnQueueForCache(unsigned char* addrRaw, int addrLen){
  int ret;
  if(debug>1) err_msg("DEBUG:=>enQueueForCache(%x,%d)", addrRaw[0],addrLen);
  ret = enQueueForCache(addrRaw,addrLen);
  if(debug>1) err_msg("DEBUG:(%d)<=enQueueForCache( )",ret);
  return ret;
}

int DeQueueForCache(unsigned char* addrRaw, int* pAddrLen){
  int ret;
  if(debug>1) err_msg("DEBUG:=>deQueueForCache( )");
  ret = deQueueForCache(addrRaw,pAddrLen);
  if(debug>1) err_msg("DEBUG:(%d)<=deQuqueForCache(%x,%d)",ret, addrRaw[0],*pAddrLen);
  return ret;

}
int ListQueueForCache(void){
  int ret;
  if(debug>1) err_msg("DEBUG:=>listQueueForCache( )");
  ret = listQueueForCache();
  if(debug>1) err_msg("DEBUG:(%d)<=listQueueForCache( )",ret);
  return ret;

}

void FreeQueueForCache(void){
  if(debug>1) err_msg("DEBUG:=>freeQueueForCache()");
  freeQueueForCache();
  if(debug>1) err_msg("DEBUG:<=freequeueForCache()");

}

int ReFormatMacAddr(char* macAddr){
  int ret;
  if(debug>1) err_msg("DEBUG:=>reFormatMacAddr(%s)", macAddr);
  ret=reFormatMacAddr(macAddr);
  if(debug>1) err_msg("DEBUG:(%d)<=reFormatMacAddr(%s)", ret, macAddr);
  return ret;
}
